import typer
from .core.errors import Errors
from .commands.manager import install_package, list_packages, uninstall_package

app = typer.Typer()

@app.command()
def install(repository: str, branch: str = "main"):
    Errors(["invalid_server"]).check()
    
    install_package(repository, branch)

@app.command()
def list():
    Errors(["invalid_server"]).check()
    
    
    for package in list_packages():
        print(package)

@app.command()
def uninstall(package: str):
    Errors(["invalid_server"]).check()
    
    uninstall_package(package)


import sys
from rich.console import Console
import requests
from tomlkit import parse

console = Console()


def fetch_toml(package: str, branch: str) -> dict:
    """
    Returns the parsed contents of a registry file from GitHub.

    Parmaters
    ---------
    package: str
        The package you want to return from.
    branch: str
        The package's branch.
    """
    author, repository = package.split("/")
    name = package_name(package, branch)

    url = (
        f"https://raw.githubusercontent.com/{author}/{repository}/{branch}/registry.toml"
    )

    response = requests.get(url)

    if not response.ok:
        error(f"Failed to fetch [red]registry.toml[/red] from [red]{name}[/red]")

    data = parse(response.text)

    if "package" not in data["tfg-kpm"]:
        error(
            'Failed to find [red]"package"[/red] section in '
            f"[red]registry.toml[/red] from [red]{name}[/red]"
        )

    return data

def package_name(package: str, branch: str) -> str:
    """
    Returns a formatted version of a package name.

    Parameters
    ----------
    package: str
        The package you want to format.
    branch: str
        The branch of the package used for formatting.
    """
    return f"{package}@{branch}"

def error(message: str) -> None:
    """
    Outputs a formatted error and stops execution.

    Parameters
    ----------
    message: str
        The message you want to output
    """

    console.print(f"[bold red]ERROR[/bold red] — [white]{message}[/white]")
    sys.exit(1)
    
    
def insert_after(lst, search_value, value):
    try:
        lst.insert(lst.index(search_value)+1, value)
    except ValueError:
        lst.append(search_value)
        
def format_strings(strings):
    formatted_list = []
    
    for s in strings:
        if any(c.isspace() for c in s):
            error("[red]Class names[/red] cannot contain [red]whitespace[/red]")
        
        s = "    " + s
        
        if not s.endswith("(event)"):
            s = s + "(event)"
        
        formatted_list.append(s)
    
    return formatted_list


from dataclasses import dataclass, field
from pathlib import Path

from .utils import error

SUPPORTED_VERSION = "2.22.0"


@dataclass
class Errors:
    """
    Handles all app errors.
    """

    errors: list[str] = field(default_factory=list[str])

    @staticmethod
    def invalid_server() -> None:
        if Path("kubejs/server_scripts/main_server_script.js").is_file():
            return

        error("Attempted to use [red]tfg-kpm[/red] command on an [red]invalid server[/red] are you sure your in the correct directory?")

    def check(self):
        for project_error in self.errors:
            getattr(self, project_error)()
            
            


from .utils import error, fetch_toml, format_strings
from dataclasses import dataclass
from typing import Self

@dataclass
class Package:
    name: str
    recipes: list[str]
    itemtags: list[str]
    blocktags: list[str]
    fluidtags: list[str]
    
    
    @classmethod
    def from_git(cls, package:str, branch:str) -> Self:
        if package.count("/") != 1:
            error("Invalid package format, expected [red]author/repo[/red]")
        
        data = fetch_toml(package, branch)
        
        if not data or "tfg-kpm" not in data or "main" not in data["tfg-kpm"] or "package" not in data["tfg-kpm"]:
            error("Invalid [red]registry.toml[/red]")
        
        main_data = data["tfg-kpm"]["main"]
        package_data = data["tfg-kpm"]["package"]
        if "name" not in main_data:
            error("Missing [red]name[/red] field in [red]registry.toml[/red]")
        # removed this to allow for event only packages
        # if "recipes" not in package_data and "itemtags" not in package_data and "blocktags" not in package_data and "fluidtags" not in package_data:
        #     error("Empty [red]package[/red] field in [red]registry.toml[/red] this package does nothing!")
        
        fields = {
            "name": main_data["name"],
            "recipes": format_strings(package_data.get("recipes", [])),
            "itemtags": format_strings(package_data.get("itemtags", [])),
            "blocktags": format_strings(package_data.get("blocktags", [])),
            "fluidtags": format_strings(package_data.get("fluidtags", []))
        }
            
        return cls(**fields)
        
        
        


from ..core.package import Package
from ..core.utils import error, insert_after, package_name, format_strings
from pathlib import Path
import requests
import sys
import io
import zipfile
import shutil
from rich.console import Console
from tomlkit import load
import os
import typer
import time


def install_package(repository: str, branch: str):
    
    console = Console()
    with console.status("Getting package metadata from github") as spinner:
        start = time.perf_counter()
        if repository.count("/") != 1:
            console.log("Failed to get package metadata", style="bold red")
            error("Invalid package format, expected [red]author/repo[/red]")
        data = Package.from_git(repository, branch)
        author, repository = repository.split("/")
        zip_url = f"https://github.com/{author}/{repository}/archive/refs/heads/{branch}.zip"
        destination = Path.cwd() / "kubejs" / "server_scripts" / "external_packages" / data.name
        
        name = package_name(repository, branch)
        console.log("Finished getting package metadata", style="bold green")
        if destination.is_dir():
            error(f"Package [red]{data.name}[/red] is already installed")
        spinner.update(status="Fetching package archive")
        destination.mkdir(parents=True, exist_ok=False)
        
        response = requests.get(zip_url)
        
        if not response.ok:
            console.log("Failed to fetch package archive", style="bold red")
            error(f"Failed to fetch [red]{name}[/red]")    
        console.log("Finished fetching package archive", style="bold green")
        spinner.update(status="Extracting package archive")
        with zipfile.ZipFile(io.BytesIO(response.content)) as z:
            for member in z.namelist():
                path = Path(member)

                # copy server_scripts contents
                if "server_scripts" in path.parts:
                    idx = path.parts.index("server_scripts")
                    relative = Path(*path.parts[idx + 1:])
                    target = destination / relative

                # copy registry.toml
                elif path.name == "registry.toml":
                    target = destination / "registry.toml"

                else:
                    continue

                if member.endswith("/"):
                    continue

                target.parent.mkdir(parents=True, exist_ok=True)
                with z.open(member) as src, target.open("wb") as dst:
                    dst.write(src.read())
        console.log("Finished extracting package to server scripts", style="bold green")
        spinner.update("Reading main_server_script.js")
        
        main_server_script = Path.cwd() / "kubejs" / "server_scripts" / "main_server_script.js"
        server_lines = []
            
        server_lines = main_server_script.read_text(encoding="utf-8").splitlines()
        console.log("Finished reading main_server_script.js", style="bold green")
        recipe_marker = "ServerEvents.recipes(event => {"
        itemtag_marker = "ServerEvents.tags('item', event => {"
        blocktag_marker = "ServerEvents.tags('block', event => {"
        fluidtag_marker = "ServerEvents.tags('fluid', event => {"
        
        spinner.update("Defining recipe classes")
        for v in data.recipes:
            insert_after(server_lines, recipe_marker, v)
            console.log(f"Finished Defining [bright_magenta]{v.replace(" ", "").replace("(event)", "")}[/bright_magenta]", style="bold green")
        console.log("Finished Defining recipe classes", style="bold green")
        spinner.update("Defining item tag classes")
        for v in data.itemtags:
            insert_after(server_lines, itemtag_marker, v)
            console.log(f"Finished Defining [bright_magenta]{v.replace(" ", "").replace("(event)", "")}[/bright_magenta]", style="bold green")
        console.log("Finished Defining item tag classes", style="bold green")
        spinner.update("Defining block tag classes")
        for v in data.blocktags:
            insert_after(server_lines, blocktag_marker, v)
            console.log(f"inished Defining [bright_magenta]{v.replace(" ", "").replace("(event)", "")}[/bright_magenta]", style="bold green")
        console.log("Finished Defining block tag classes", style="bold green")
        spinner.update("Defining fluid tag classes")
        for v in data.fluidtags:
            insert_after(server_lines, fluidtag_marker, v)
            console.log(f"Finished Defining [bright_magenta]{v.replace(" ", "").replace("(event)", "")}[/bright_magenta]", style="bold green")
        console.log("Finished Defining fluid tag classes", style="bold green")
        spinner.update("Writing to main_server_script.js")
        with open(main_server_script, "w", encoding="utf-8") as f:
            for item in server_lines:
                f.write(f"{item}\n")
    console.log("Finished writing to main_server_script.js", style="bold green")
    end = time.perf_counter()
    elapsed = end - start

    if elapsed < 1:
        console.log(f"Installed {data.name} in [bright_cyan]{elapsed * 1000:.0f}[/bright_cyan] milliseconds", style="bold green")
    elif elapsed < 60:
        console.log(f"Installed {data.name} in [bright_cyan]{elapsed:.2f}[/bright_cyan] seconds", style="bold green")
    else:
        minutes = elapsed / 60
        console.log(f"Installed {data.name} in [bright_cyan]{minutes:.2f}[/bright_cyan] minutes", style="bold green")
        
def list_packages():
    packagefolder = Path.cwd() / "kubejs" / "server_scripts" / "external_packages"
    if not packagefolder.is_dir():
        print("No packages installed")
        sys.exit(0)
    packages = os.listdir(packagefolder)
    if not packages:
        print("No packages installed")
        sys.exit(0)
    return packages
    
def uninstall_package(package: str):
    console = Console()
    
    installedpackages = list_packages()
    if package not in installedpackages:
        error(f"{package} is [red]not[/red] installed!")
    
    packagedir = Path.cwd() / "kubejs" / "server_scripts" / "external_packages" / package
        
    console.log(f"Found existing installation: {package}", style="bold green")
    console.log(f"Uninstalling {package}:")
    console.log("  Would remove:")
    console.log(f"    [bright_cyan]{packagedir.as_posix()}[/bright_cyan]")
    typer.confirm("Proceed?", abort=True)
    start = time.perf_counter()
    with console.status("Parsing registry file") as spinner:
        registry = packagedir / "registry.toml"
        with open(registry, "r", encoding="utf-8") as f:
            data = load(f)
        package_data = data["tfg-kpm"]["package"]
        classes = package_data.get("recipes") + package_data.get("itemtags") + package_data.get("blocktags") + package_data.get("fluidtags")
        classes = format_strings(classes)
        console.log("Finished parsing registry file", style="bold green")
        spinner.update(status="Reading main_server_script.js")
        main_server_script = Path.cwd() / "kubejs" / "server_scripts" / "main_server_script.js"
        
        server_lines = main_server_script.read_text(encoding="utf-8").splitlines()
        
        console.log("Finished reading main_server_script.js", style="bold green")
        spinner.update(status="Removing class definitions")

        comparisonset = set(classes)
        server_lines = [item for item in server_lines if item not in comparisonset]
        
        console.log("Finished removing class definitions", style="bold green")
        spinner.update(status="Writing to main_server_script.js")
        
        with open(main_server_script, "w", encoding="utf-8") as f:
                for item in server_lines:
                    f.write(f"{item}\n")
        
        console.log("Finished writing to main_server_script.js", style="bold green")
        spinner.update(status="Removing package")
        
        shutil.rmtree(packagedir)
        
        console.log("Finished removing package", style="bold green")
    end = time.perf_counter()
    elapsed = end - start

    if elapsed < 1:
        console.log(f"Uninstalled {package} in [bright_cyan]{elapsed * 1000:.0f}[/bright_cyan] milliseconds", style="bold green")
    elif elapsed < 60:
        console.log(f"Uninstalled {package} in [bright_cyan]{elapsed:.2f}[/bright_cyan] seconds", style="bold green")
    else:
        minutes = elapsed / 60
        console.log(f"Uninstalled {package} in [bright_cyan]{minutes:.2f}[/bright_cyan] minutes", style="bold green")
    
    
    